/*===========================================================================*
 - Copyright 2010 Google Inc.
 -
 - Licensed under the Apache License, Version 2.0 (the "License");
 - you may not use this file except in compliance with the License.
 - You may obtain a copy of the License at
 -
 -     http://www.apache.org/licenses/LICENSE-2.0
 -
 - Unless required by applicable law or agreed to in writing, software
 - distributed under the License is distributed on an "AS IS" BASIS,
 - WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 - See the License for the specific language governing permissions and
 - limitations under the License.
 *===========================================================================*/

/*
 * webpconv.c
 *
 *   Converts an image file from one format to another.
 *
 *   Usage: webpconv [OPTION] input_file(s)
 *          Converts the input_file(s) to vp8 or other format
 *
 *   -output_dir           output directory
 *   -format               output image format
 *   -NQP                  parameter to trade off between the size and quality
 *                         of output image. Should be within 0-100. Higher NQP
 *                         means bigger and better image.
 *
 *      where format is one of these:
 *
 *         BMP
 *         JPEG  (only applicable for 8 bpp or rgb)
 *         PNG
 *         TIFF
 *         TIFF_G4  (only applicable for 1 bpp)
 *         PNM
 *         VP8
 *
 *   The output format can be chosen by the extension of fileout:
 *         BMP       .bmp
 *         JPEG      .jpg
 *         PNG       .png
 *         TIFF      .tif
 *         TIFF_G4   .tif
 *         PNM       .pnm
 *         VP8
 */

#ifdef __cplusplus
extern "C" {
#endif  /* __cplusplus */

#include <ctype.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "allheaders.h"

#define DEFAULT_PSNR 42.0
#define MIN_NQP 0
#define MAX_NQP 100
#define MIN_VP8QP 1
#define MAX_VP8QP 63

static char kValidFormats[] = "Valid formats: BMP, JPEG, PNG, TIFF, TIFF_G4, PNM, VP8";

int isdir(const char *in_string)
{
  DIR *dir;
  dir = opendir(in_string);
  if (dir != NULL)
  {
    closedir(dir);
    return 1;
  }

  return 0;
}

void usage() {
  fprintf(stderr, "Usage: webpconv [OPTION] input_file(s)\n"
          "Converts the input_file(s) to vp8 or other format\n\n"
          "  -output_dir        output directory\n"
          "  -format            output image format\n"
          "  -NQP               parameter to trade off between size and\n"
          "                     quality of the output image. NQP should be\n"
          "                     within 0-100. Higher NQP means bigger and better image\n");
  fprintf(stderr, "\n%s\n", kValidFormats);
  fprintf(stderr, "\nExamples:\n"
          "webpconv /home/user/images/img.png\n"
          "Will create /home/user/images/img.vp8\n\n"
          "webpconv -output_dir /tmp -format PNG /home/user/images/*.vp8\n"
          "Will create /tmp/*.png images\n");
}

/*
 * Maps normalized QP to VP8 QP
 */
int mapNQPToVP8QP(int NQP) {
  if (NQP < MIN_NQP || NQP > MAX_NQP) {
    fprintf(stderr, "Wrong NQP value %d.", NQP);
    return -1;
  }

  float scale = MAX_VP8QP - MIN_VP8QP;
  float vp8qp = scale * (MAX_NQP - NQP) / (MAX_NQP - MIN_NQP) + MIN_VP8QP;
  return (int)vp8qp;
}

/*
 * Maps VP8 QP to normalized QP
 */
int mapVP8QPToNQP(int VP8QP) {
  if (VP8QP < MIN_VP8QP || VP8QP > MAX_VP8QP) {
    fprintf(stderr, "Wrong VP8QP value %d.", VP8QP);
    return -1;
  }

  float scale = MAX_NQP - MIN_NQP;
  float nqp = scale * (MAX_VP8QP - VP8QP) / (MAX_VP8QP - MIN_VP8QP) + MIN_NQP;
  return (int)nqp;
}

int main(int    argc,
         char **argv)
{
  char *out_dir_arg = NULL;
  const char *format = "vp8";
  int NQP = -1;

  if (argc < 2) {
    usage();
    return 1;
  }

  int parse_failed = FALSE;
  int a_idx = 1;
  for (; (a_idx < argc) && (argv[a_idx][0] == '-'); a_idx++) {
    if (strcmp(argv[a_idx], "-output_dir") == 0) {
      out_dir_arg = argv[++a_idx];
      printf("Output directory %s\n", out_dir_arg);

      if (!isdir(out_dir_arg)) {
        fprintf(stderr, "unable to open %s directory\n", out_dir_arg);
        parse_failed = TRUE;
      }

      if (out_dir_arg[strlen(out_dir_arg) - 1] != '/') {
        out_dir_arg = stringJoin(out_dir_arg, "/");
      }
    } else if (strcmp(argv[a_idx], "-format") == 0) {
      a_idx++;
      int j = 0;
      for (; argv[a_idx][j]; j++) {
        argv[a_idx][j] = tolower(argv[a_idx][j]);
      }
      format = argv[a_idx];
      printf("Output format %s\n", format);
    } else if (strcmp(argv[a_idx], "-NQP") == 0) {
      NQP = atoi(argv[++a_idx]);
      if (NQP < MIN_NQP || NQP > MAX_NQP) {
        fprintf(stderr, "Wrong NQP value %d. Should be within 0-100", NQP);
        parse_failed = TRUE;
      }

      printf("Target NQP %d\n", NQP);
    } else {
      fprintf(stderr, "Wrong command-line parameter %s\n\n", argv[a_idx]);
      parse_failed = TRUE;
    }
  }

  if (parse_failed) {
    usage();
    return 1;
  }

  if (a_idx >= argc) {
    fprintf(stderr, "No input file specified\n");
    usage();
  }

  for(; a_idx < argc; a_idx++)
  {
    if (isdir(argv[a_idx])) {
      continue;
    }

    printf("processing %s\n", argv[a_idx]);
    char* out_dir = out_dir_arg;
    if (out_dir == NULL) {
      splitPathAtDirectory(argv[a_idx], &out_dir, NULL);
    }

    struct Pix *pixs;
    char* raw_name, *ext;
    splitPathAtExtension(argv[a_idx], &raw_name, &ext);
    if (!strcmp(ext, ".vp8")) {
      pixs = pixReadVp8(argv[a_idx]);
    } else {
      pixs = pixRead(argv[a_idx]);
    }

    char* file_name, *base;
    splitPathAtDirectory(raw_name, NULL, &file_name);
    base = stringJoin(out_dir, file_name);
    base = stringJoin(base, ".");

    l_int32 d = pixGetDepth(pixs);
    if (d != 1 && !strcmp(format, "TIFF_G4")) {
      L_WARNING("can't convert to tiff_g4; converting to tiff", "vp8conv");
      format = "tiff";
    }
    if (d < 8 && !strcmp(format, "JPEG")) {
      L_WARNING("can't convert to jpeg; converting to png", "vp8conv");
      format = "png";
    }

    const char* fileout = stringJoin(base, format);
    fprintf(stderr, "Output file %s\n", fileout);
    if (strcmp(format, "bmp") == 0) {
      pixWrite(fileout, pixs, IFF_BMP);
    } else if (strcmp(format, "jpeg") == 0) {
      pixWrite(fileout, pixs, IFF_JFIF_JPEG);
    } else if (strcmp(format, "png") == 0) {
      pixWrite(fileout, pixs, IFF_PNG);
    } else if (strcmp(format, "tiff") == 0) {
      pixWrite(fileout, pixs, IFF_TIFF_ZIP);
    } else if (strcmp(format, "tiff_g4") == 0) {
      pixWrite(fileout, pixs, IFF_TIFF_G4);
    } else if (strcmp(format, "pnm") == 0) {
      pixWrite(fileout, pixs, IFF_PNM);
    } else if (strcmp(format, "vp8") == 0) {
      if (NQP >= 0) {
        int vp8qp = mapNQPToVP8QP(NQP);
        pixWriteVp8(fileout, pixs, vp8qp);
      } else {
        int vp8qp = 0;
        pixWriteVp8withTargetPSNR(fileout, pixs, DEFAULT_PSNR, &vp8qp);

        printf("Used NQP=%d\n", mapVP8QPToNQP(vp8qp));
      }
    }
    else {
      return ERROR_INT(kValidFormats, mainName, 1);
    }
  }

  return 0;
}
