/*===========================================================================*
 - Copyright 2010 Google Inc.
 -
 - This code is licensed under the same terms as WebM:
 - Software License Agreement:  http://www.webmproject.org/license/software/
 - Additional IP Rights Grant:  http://www.webmproject.org/license/additional/
 *===========================================================================*/

/*
 * webpconv.c
 *
 * Converts an image file from one format to another. Presently conversion
 * between the following formats are supported.
 *
 *      BMP
 *      JPEG  (only applicable for 8 bpp or rgb)
 *      PNG
 *      WEBP
 *
 * Other image formats (e.g. TIFF) can be supported if leptonica is compiled
 * with the support of corresponding formats. See leptonica README
 * (http://www.leptonica.com/source/README.html) to know how to enable support
 * for other image formats.
 *
 *
 * Notes on the 'quality' command line parameter
 *
 * The parameter used in WebP encoder to trade off between image quality and
 * size is QP (not 'quality'). 'quality' parameter here denotes normalized QP.
 * The meaning and range of QP is quite different from the quality parameters
 * used for other image formats (e.g. the range of QP is between 1 to 63 and
 * lower QP value means better image). Therefore we normalize the QP such that
 * higher value of 'quality'  means better image and also it has a range between
 * 0 to 100.
 */

#include <ctype.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "allheaders.h"

#define DEFAULT_PSNR 42.0
#define MIN_QUALITY 0
#define MAX_QUALITY 100
#define MIN_VP8QP 1
#define MAX_VP8QP 63

int isdir(const char *in_string)
{
  DIR *dir;
  dir = opendir(in_string);
  if (dir != NULL)
  {
    closedir(dir);
    return 1;
  }

  return 0;
}

void usage() {
  fprintf(stderr, "Usage: webpconv [-output_dir DIR] [-format FORMAT] "
          "[-quality 0..100] input_file(s)\n"
          "Converts the input_file(s) to WebP or other format\n\n"
          "  -output_dir DIR     specifies output directory. [default: input\n"
          "                      image directory].\n"
          "  -format FORMAT      specifies output image format. Valid formats "
          "are:\n"
          "                      BMP, JPEG, PNG, WEBP. [default WEBP].\n"
          "  -quality 0..100     parameter to trade off between size and\n"
          "                      quality of the output WebP image. quality is\n"
          "                      set from 0 to 100. A higher quality value\n"
          "                      means bigger and better image.\n"
          "                      Note that this parameter is applicable only\n"
          "                      when the output format is WEBP.");
  fprintf(stderr, "\nExamples:\n"
          "webpconv /home/user/images/img.png\n"
          "Will create /home/user/images/img.webp\n\n"
          "webpconv -output_dir /tmp -format PNG /home/user/images/*.webp\n"
          "Will create /tmp/*.png images\n");
}

/*
 * Maps normalized QP (qaulity) to VP8 QP
 */
int mapQualityToVP8QP(int quality) {
  if (quality < MIN_QUALITY || quality > MAX_QUALITY) {
    fprintf(stderr, "Wrong quality value %d.", quality);
    return -1;
  }

  const float scale = MAX_VP8QP - MIN_VP8QP;
  const float vp8qp =
      scale * (MAX_QUALITY - quality) / (MAX_QUALITY - MIN_QUALITY) + MIN_VP8QP;
  return (int)(vp8qp + 0.5);
}

/*
 * Maps VP8 QP to normalized QP (quality)
 */
int mapVP8QPToQuality(int VP8QP) {
  if (VP8QP < MIN_VP8QP || VP8QP > MAX_VP8QP) {
    fprintf(stderr, "Wrong VP8QP value %d.", VP8QP);
    return -1;
  }

  const float scale = MAX_QUALITY - MIN_QUALITY;
  const float nqp = scale * (MAX_VP8QP - VP8QP) / (MAX_VP8QP - MIN_VP8QP)
      + MIN_QUALITY;
  return (int)(nqp + 0.5);
}

int main(int    argc,
         char **argv)
{
  char *out_dir_arg = NULL;
  const char *format = "webp";
  int quality = -1;

  if (argc < 2) {
    usage();
    return 1;
  }

  int parse_failed = FALSE;
  int a_idx = 1;
  for (; (a_idx < argc) && (argv[a_idx][0] == '-'); a_idx++) {
    if (strcmp(argv[a_idx], "-output_dir") == 0) {
      out_dir_arg = argv[++a_idx];
      printf("Output directory %s\n", out_dir_arg);

      if (!isdir(out_dir_arg)) {
        fprintf(stderr, "unable to open %s directory\n", out_dir_arg);
        parse_failed = TRUE;
      }

      if (out_dir_arg[strlen(out_dir_arg) - 1] != '/') {
        out_dir_arg = stringJoin(out_dir_arg, "/");
      }
    } else if (strcmp(argv[a_idx], "-format") == 0) {
      a_idx++;
      int j = 0;
      for (; argv[a_idx][j]; j++) {
        argv[a_idx][j] = tolower(argv[a_idx][j]);
      }
      format = argv[a_idx];
      printf("Output format %s\n", format);
    } else if (strcmp(argv[a_idx], "-quality") == 0) {
      quality = atoi(argv[++a_idx]);
      if (quality < MIN_QUALITY || quality > MAX_QUALITY) {
        fprintf(stderr, "Wrong quality value %d. Should be within 0-100\n",
                quality);
        parse_failed = TRUE;
      }

      printf("Target quality %d\n", quality);
    } else {
      fprintf(stderr, "Wrong command-line parameter %s\n\n", argv[a_idx]);
      parse_failed = TRUE;
    }
  }

  if (parse_failed) {
    usage();
    return 1;
  }

  if (a_idx >= argc) {
    fprintf(stderr, "No input file specified\n");
    usage();
  }

  for (; a_idx < argc; a_idx++) {
    if (isdir(argv[a_idx])) {
      continue;
    }

    printf("processing %s\n", argv[a_idx]);
    char *out_dir = out_dir_arg;
    if (out_dir == NULL) {
      splitPathAtDirectory(argv[a_idx], &out_dir, NULL);
    }

    char *raw_name, *ext;
    char *file_name;

    splitPathAtExtension(argv[a_idx], &raw_name, &ext);
    splitPathAtDirectory(raw_name, NULL, &file_name);
    const int out_dir_len = strlen(out_dir);
    const int file_name_len = strlen(file_name);
    char* fileout = (char *)calloc(
        (out_dir_len + file_name_len + strlen(format)) + 2, sizeof(char));
    if (fileout == NULL) {
      fprintf(stderr, "calloc failed for fileout");
      return 1;
    }

    if (!strcmp(format, "jpeg")) {
      sprintf(fileout, "%s%s.%s", out_dir, file_name, "jpg");
    } else {
      sprintf(fileout, "%s%s.%s", out_dir, file_name, format);
    }
    printf("Output file %s\n", fileout);

    FREE(raw_name);
    FREE(file_name);

    FILE *fp = fopen(fileout, "r");
    if (fp)
    {
      fclose(fp);
      free(fileout);
      FREE(ext);
      fprintf(stderr, "Output file exists. You may change the output_dir\n");
      continue;
    }

    struct Pix *pixs;
    if (!strcmp(ext, ".webp")) {
      pixs = pixReadWebP(argv[a_idx]);
    } else {
      pixs = pixRead(argv[a_idx]);
    }
    FREE(ext);

    if (pixs == NULL) {
      fprintf(stderr, "Failed to read image\n");
      return 1;
    }

    l_int32 d = pixGetDepth(pixs);
    if (d != 1 && !strcmp(format, "tiff_g4")) {
      L_WARNING("can't convert to tiff_g4; converting to tiff", "webpconv");
      format = "tiff";
    }
    if (d < 8 && !strcmp(format, "jpeg")) {
      L_WARNING("can't convert to jpeg; converting to png", "webpconv");
      format = "png";
    }

    if (strcmp(format, "bmp") == 0) {
      pixWrite(fileout, pixs, IFF_BMP);
    } else if (strcmp(format, "jpeg") == 0) {
      pixWrite(fileout, pixs, IFF_JFIF_JPEG);
    } else if (strcmp(format, "png") == 0) {
      pixWrite(fileout, pixs, IFF_PNG);
    } else if (strcmp(format, "tiff") == 0) {
      pixWrite(fileout, pixs, IFF_TIFF_ZIP);
    } else if (strcmp(format, "tiff_g4") == 0) {
      pixWrite(fileout, pixs, IFF_TIFF_G4);
    } else if (strcmp(format, "pnm") == 0) {
      pixWrite(fileout, pixs, IFF_PNM);
    } else if (strcmp(format, "webp") == 0) {
      if (quality >= 0) {
        int vp8qp = mapQualityToVP8QP(quality);
        if (pixWriteWebP(fileout, pixs, vp8qp) != 0) {
          fprintf(stderr, "pixWriteWebP Failed. See previous error messages. "
                  "Turn on debug build if you aren't seeing enough detailed "
                  "error messages\n");
        }
      } else {
        int vp8qp = 0;
        if (pixWriteWebPwithTargetPSNR(fileout, pixs, DEFAULT_PSNR, &vp8qp)
            == 0) {
          printf("Used quality=%d\n", mapVP8QPToQuality(vp8qp));
        } else {
          fprintf(stderr, "pixWriteWebPwithTargetPSNR Failed. See previous "
                  "error messages. Turn on debug build if you aren't "
                  "seeing enough detailed error messages\n");
        }
      }
    }
    else {
      return ERROR_INT("Format not supported", "webpconv", 1);
    }

    free(fileout);
    pixDestroy(&pixs);
  }

  return 0;
}
